# -*- encoding: utf-8 -*-
"""
hio.base.filing module

"""
import os
import stat
import shutil
import tempfile
from contextlib import contextmanager


from . import doing
from .. import hioing
from .. import help
from ..help.helping import ocfn

logger = help.ogler.getLogger()


class Filer(hioing.Mixin):
    """
    Filer instances manage file directories and files to hold keri installation
    specific resources like databases and configuration files.

    Class Attributes:
        HeadDirPath (str): default abs dir path head such as "/usr/local/var"
        TailDirPath (str): default rel dir path tail when using head
        CleanTailDirPath (str): default rel dir path tail when creating clean
        AltHeadDirPath (str): default alt dir path head such as  "~"
                              as fallback when desired head not permitted.
        AltTailDirPath (str): default alt rel dir path tail as fallback
                              when using alt head.
        AltCleanTailDirPath (str): default alt rel path tail when creating clean
        TempHeadDir (str): default temp abs dir path head such as "/tmp"
        TempPrefix (str): default rel dir path prefix when using temp head
        TempSuffix (str): default rel dir path suffix when using temp head and tail
        Perm (int): explicit default octal perms such as 0o1700
        Mode (str): open mode such as "r+"
        Fext (str): default file extension such as "text" for "fname.text"


    Attributes:
        name (str): unique path component used in directory or file path name
        base (str): another unique path component inserted before name
        temp (bool): True means use TempHeadDir in /tmp directory
        headDirPath (str): head directory path
        path (str | None):  full directory or file path once created else None
        perm (int):  octal OS permissions for path directory and/or file
        filed (bool): True means .path ends in file.
                       False means .path ends in directory
        extensioned (bool): When not filed:
                                True means ensure .path ends with fext
                                False means do not ensure .path ends with fext
        mode (str): file open mode if filed
        fext (str): file extension if filed
        file (File | None): File instance when filed and created.
        opened (bool): True means directory created and if filed then file
                is opened. False otherwise


    File/Directory Creation Mode Notes:
        .Perm provides default restricted access permissions to directory and/or files
        stat.S_ISVTX | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
        0o1700==960

        stat.S_ISVTX  is Sticky bit. When this bit is set on a directory it means
            that a file in that directory can be renamed or deleted only by the
            owner of the file, by the owner of the directory, or by a privileged process.
            When this bit is set on a file it means nothing
        stat.S_IRUSR Owner has read permission.
        stat.S_IWUSR Owner has write permission.
        stat.S_IXUSR Owner has execute permission.
    """
    HeadDirPath = os.path.join(os.path.sep, 'usr', 'local', 'var')  # default in /usr/local/var
    TailDirPath = "hio"
    CleanTailDirPath = os.path.join("hio", "clean")


    AltHeadDirPath = os.path.expanduser("~")  # put in ~ as fallback when desired not permitted
    AltTailDirPath = ".hio"
    AltCleanTailDirPath = os.path.join(".hio", "clean")
    TempHeadDir = os.path.join(os.path.sep, "tmp")
    TempPrefix = "hio_"
    TempSuffix = "_test"
    Perm = stat.S_ISVTX | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR  # 0o1700==960
    Mode = "r+"
    Fext = "text"

    def __init__(self, *, name='main', base="", temp=False, headDirPath=None,
                 perm=None, reopen=True, clear=False, reuse=False, clean=False,
                 filed=False, extensioned=False, mode=None, fext=None, **kwa):
        """Setup directory of file at .path

        Parameters:
            name (str): Unique identifier of file. Unique directory path name
                differentiator directory/file
                When system employs more than one keri installation, name allows
                differentiating each instance by name
            base (str): optional directory path segment inserted before name
                that allows further differentation with a hierarchy. "" means
                optional.
            temp (bool): assign to .temp
                True then open in temporary directory, clear on close
                Otherwise then open persistent directory, do not clear on close
            headDirPath (str): optional head directory pathname for main database
                Default .HeadDirPath
            perm (int): optional numeric os dir permissions for database
                directory and database files. Default .DirMode
            reopen (bool): True (re)open with this init
                           False not (re)open with this init but later (default)
            clear (bool): True means remove directory upon close when reopening
                          False means do not remove directory upon close when reopening
            reuse (bool): True means reuse self.path if already exists
                          False means do not reuse but remake self.path
            clean (bool): True means path uses clean tail variant
                             False means path uses normal tail variant
            filed (bool): True means .path is file path not directory path
                          False means .path is directiory path not file path
            extensioned (bool): When not filed:
                                True means ensure .path ends with fext
                                False means do not ensure .path ends with fext
            mode (str): File open mode when filed
            fext (str): File extension when filed or extensioned

        """
        super(Filer, self).__init__(**kwa)  # Mixin for Mult-inheritance MRO

        # ensure relative path parts are relative because of odd path.join behavior
        if os.path.isabs(name):
            raise hioing.FilerError(f"Not relative {name=} path.")
        if os.path.isabs(base):
            raise hioing.FilerError(f"Not relative {base=} path.")

        self.name = name
        self.base = base
        self.temp = True if temp else False
        self.headDirPath = headDirPath if headDirPath is not None else self.HeadDirPath
        self.perm = perm if perm is not None else self.Perm
        self.path = None
        self.filed = True if filed else False
        self.extensioned = True if extensioned else False
        self.mode = mode if mode is not None else self.Mode
        self.fext = fext if fext is not None else self.Fext
        self.file = None
        self.opened = False

        if reopen:
            self.reopen(clear=clear, reuse=reuse, clean=clean, **kwa)


    def reopen(self, temp=None, headDirPath=None, perm=None, clear=False,
               reuse=False, clean=False, mode=None, fext=None, **kwa):
        """
        Open if closed or close and reopen if opened or create and open if not

        Parameters:
            temp (bool): assign to .temp
                         True means open in temporary directory, clear on close
                         False means open persistent directory, do not clear on close
            headDirPath (str): optional head directory pathname for main database
                               Default .HeadDirpath
            perm (int): optional numeric os dir permissions for database
                         directory and database files. Default .Perm
            clear (bool): True means remove directory upon close
                             False means do not remove directory upon close
            reuse (bool): True means reuse self.path if already exists
                             False means do not reuse but remake self.path
            clean (bool): True means path uses clean tail variant
                             False means path uses normal tail variant
            mode (str): file open mode when .filed
            fext (str): File extension when .filed
        """
        self.close(clear=clear)

        if temp is not None:
            self.temp = temp
        if headDirPath is not None:
            self.headDirPath = headDirPath
        if perm is not None:
            self.perm = perm
        if mode is not None:
            self.mode = mode
        if fext is not None:
            self.fext = fext

        if not self.path or not os.path.exists(self.path) or not reuse:
            self.path, self.file = self.remake(name=self.name,
                                               base=self.base,
                                               temp=self.temp,
                                               headDirPath=self.headDirPath,
                                               perm=self.perm,
                                               clean=clean,
                                               filed=self.filed,
                                               extensioned=self.extensioned,
                                               mode=self.mode,
                                               fext=self.fext,
                                               **kwa)
        elif self.filed:  # assumes dir in self.path exists
            self.file = ocfn(self.path, mode=self.mode)

        self.opened = True if not self.filed else self.file and not self.file.closed

        return self.opened


    def remake(self, *, name="", base="", temp=None, headDirPath=None, perm=None,
                clean=False, filed=False, extensioned=False, mode=None,
                fext=None, **kwa):
        """
        Make and return (path. file) by opening or creating and opening if not
        preexistent, directory or file at  path

        Parameters:
            name (str): unique name alias portion of path
            base (str): optional base inserted before name in path
            temp (bool): optional
                None means ignore,
                True means open temporary directory, may clear on close
                False means open persistent directory, may not clear on close

            headDirPath (str): optional head directory pathname of main database

            perm (int): directory or file permissions such as
                stat.S_IRUSR Owner has read permission.
                stat.S_IWUSR Owner has write permission.
                stat.S_IXUSR Owner has execute permission.

            clean (bool): True means make path for cleaned version and  remove
                          old directory or file at clean path if any.
                          False means make path normally (not clean)

            filed (bool): True means .path is file path not directory path
                          False means .path is directiory path not file path
            extensioned (bool): When not filed:
                                True means ensure .path ends with fext
                                False means do not ensure .path ends with fext
            mode (str): file open mode when .filed such as "w+"
            fext (str): File extension when .filed
        """
        if os.path.isabs(name):
            raise hioing.FilerError(f"Not relative {name=} path.")

        if os.path.isabs(base):
            raise hioing.FilerError(f"Not relative {base=} path.")

        file = None
        temp = True if temp else False

        # use class defaults here so can use makePath for other dirs and files
        if headDirPath is None:
            headDirPath = self.HeadDirPath
        if perm is None:
            perm = self.Perm
        if mode is None:
            mode = self.Mode
        if fext is None:
            fext = self.Fext

        tailDirPath = self.CleanTailDirPath if clean else self.TailDirPath
        altTailDirPath = self.AltCleanTailDirPath if clean else self.AltTailDirPath

        if filed or extensioned:
            root, ext = os.path.splitext(name)
            if not ext:
                name = f"{name}.{fext}"

        if os.path.isabs(name):
            raise hioing.FilerError(f"Not relative {name=} path.")

        if temp:
            headDirPath = tempfile.mkdtemp(prefix=self.TempPrefix,
                                           suffix=self.TempSuffix,
                                           dir=self.TempHeadDir)

            path = os.path.abspath(
                                os.path.join(headDirPath,
                                             tailDirPath,
                                             base,
                                             name))

            if clean and os.path.exists(path):
                if os.path.isfile(path):
                    if filed or extensioned:
                        os.remove(path)  # rm only file not dir
                    else:
                        head, tail = os.path.split(path)
                        shutil.rmtree(head)  # rm directory and all files
                else:
                    shutil.rmtree(path)

            if filed or extensioned:
                head, tail = os.path.split(path)
                if not os.path.exists(head):
                    os.makedirs(head)
                if filed:
                    file = ocfn(path, mode=mode, perm=perm)
            else:
                os.makedirs(path)

        else:
            path = os.path.abspath(
                        os.path.expanduser(
                            os.path.join(headDirPath,
                                         tailDirPath,
                                         base,
                                         name)))

            if clean and os.path.exists(path):
                if os.path.isfile(path):
                    if filed:
                        os.remove(path)  # rm only file not dir
                    else:
                        head, tail = os.path.split(path)
                        shutil.rmtree(head)  # rm directory and all files
                else:
                    shutil.rmtree(path)

            if not os.path.exists(path):  # no path so attempt to create
                try:
                    if filed or extensioned:
                        head, tail = os.path.split(path)
                        if not os.path.exists(head):
                            os.makedirs(head)
                        if filed:
                            file = ocfn(path, mode=mode, perm=perm)
                    else:
                        os.makedirs(path)

                except OSError as ex:  # use alt instead should always succeed
                    headDirPath = self.AltHeadDirPath
                    path = os.path.abspath(
                                os.path.expanduser(
                                    os.path.join(headDirPath,
                                                 altTailDirPath,
                                                 base,
                                                 name)))
                    if not os.path.exists(path):
                        if filed or extensioned:
                            head, tail = os.path.split(path)
                            if not os.path.exists(head):
                                os.makedirs(head)
                            if filed:
                                file = ocfn(path, mode=mode, perm=perm)

                        else:
                            os.makedirs(path)
                    else:
                        if filed:
                            file = ocfn(path, mode=mode, perm=perm)

            else:  # verify access
                if not os.access(path, os.F_OK | os.R_OK | os.W_OK): # use alt instead
                    headDirPath = self.AltHeadDirPath
                    path = os.path.abspath(
                                os.path.expanduser(
                                    os.path.join(headDirPath,
                                                 altTailDirPath,
                                                 base,
                                                 name)))
                    if not os.path.exists(path):
                        if filed or extensioned:
                            head, tail = os.path.split(path)
                            if not os.path.exists(head):
                                os.makedirs(head)
                            if filed:
                                file = ocfn(path, mode=mode, perm=perm)

                        else:
                            os.makedirs(path)
                else:
                    if filed:
                        file = ocfn(path, mode=mode, perm=perm)

            if not extensioned:
                os.chmod(path, perm)  # set dir/file permissions

        return path, file


    def exists(self, name="", base="", headDirPath=None, clean=False,
               filed=False, extensioned=False, fext=None):
        """
        Check if (path. file) exists for a given set of parameters for remake.
        Temp is not allowed.

        Parameters:
            name (str): unique name alias portion of path
            base (str): optional base inserted before name in path

            headDirPath (str): optional head directory pathname of main database

            clean (bool): True means make path for cleaned version and  remove
                          old directory or file at clean path if any.
                          False means make path normally (not clean)

            filed (bool): True means .path is file path not directory path
                          False means .path is directiory path not file path
            extensioned (bool): When not filed:
                                True means ensure .path ends with fext
                                False means do not ensure .path ends with fext
            fext (str): File extension when .filed

        Returns:
            bool: True means path or alt path exists, false means neither exists

        """
        if os.path.isabs(name):
            raise hioing.FilerError(f"Not relative {name=} path.")

        if os.path.isabs(base):
            raise hioing.FilerError(f"Not relative {base=} path.")

        # use class defaults here so can use makePath for other dirs and files
        if headDirPath is None:
            headDirPath = self.HeadDirPath

        if fext is None:
            fext = self.Fext

        tailDirPath = self.CleanTailDirPath if clean else self.TailDirPath
        altTailDirPath = self.AltCleanTailDirPath if clean else self.AltTailDirPath

        if filed or extensioned:
            root, ext = os.path.splitext(name)
            if not ext:
                name = f"{name}.{fext}"

        if os.path.isabs(name):
            raise hioing.FilerError(f"Not relative {name=} path.")

        path = os.path.abspath(
            os.path.expanduser(
                os.path.join(headDirPath,
                             tailDirPath,
                             base,
                             name)))

        # Check non-alt, if exists return True
        if os.path.exists(path):
            return True

        # Now we must check the alt path to see if that exists.
        headDirPath = self.AltHeadDirPath
        path = os.path.abspath(
                os.path.expanduser(
                    os.path.join(headDirPath,
                                altTailDirPath,
                                base,
                                name)))
        return os.path.exists(path)


    def close(self, clear=False):
        """
        Close .file if any and if clear rm directory or file at .path

        Parameters:
           clear (bool): True means remove dir or file at .path
        """
        if self.file:
            self.file.close()
        self.opened = False

        if clear:
            self._clearPath()

        return not self.opened  # True means closed False means still opened


    def _clearPath(self):
        """
        Remove directory/file at end of .path
        """
        if self.path and os.path.exists(self.path):
            if os.path.isfile(self.path):  # self.path points to File instance
                os.remove(self.path)  # rm only file at end of path

                if self.file:  # self.file is File instance
                    self.file = None  #

                if self.temp:  # remove trailing dir of path as well
                    head, tail = os.path.split(self.path)
                    shutil.rmtree(head)  # rm dir head as root and all below head

            elif self.extensioned:  # path end has file extension
                os.remove(self.path)  # rm only extensioned end of path

                if self.temp:  # remove trailing dir of path as well
                    head, tail = os.path.split(self.path)
                    shutil.rmtree(head)  # rm dir head as root and all below head

            else:
                shutil.rmtree(self.path)  # remove trailing dir of path (and all below)



@contextmanager
def openFiler(cls=None, name="test", temp=True, reopen=True, clear=False, **kwa):
    """
    Context manager wrapper Filer instances for managing a filesystem directory
    and or files in a directory.

    Defaults to using temporary directory path.
    Context 'with' statements call .close on exit of 'with' block

    Parameters:
        cls is Class instance of subclass instance
        name is str name of Filer instance path part so can have multiple Filers
             at different paths that each use different dirs or files
        temp is Boolean, True means open in temporary directory, clear on close
                Otherwise open in persistent directory, do not clear on close
        reopen (bool): True (re)open with this init
                           False not (re)open with this init but later (default)
        clear (bool): True means remove directory upon close when reopening
                          False means do not remove directory upon close when reopening
    See filing.Filer for other keyword parameter passthroughs

    Usage:

    with openFiler(name="bob") as filer:

    with openFiler(name="eve", cls=FilerSubClass) as filer:

    """
    filer = None
    if cls is None:
        cls = Filer
    try:
        filer = cls(name=name, temp=temp, reopen=reopen, clear=clear, **kwa)
        yield filer

    finally:
        if filer:
            filer.close(clear=filer.temp or clear)  # clears if filer.temp



class FilerDoer(doing.Doer):
    """
    Basic Filer Doer

    Attributes:
        done (bool): completion state:
            True means completed
            Otherwise incomplete. Incompletion maybe due to close or abort.
        filer (Filer): instance

    Properties:
        tyme (float): relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        tymth (func): closure returned by Tymist .tymeth() method.
            When .tymth is called it returns associated Tymist .tyme.
            .tymth provides injected dependency on Tymist tyme base.
        tock (float)): desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    """

    def __init__(self, filer, **kwa):
        """
        Parameters:
           tymist (Tymist): instance
           tock (float): initial value of .tock in seconds
           filer (Filer): instance
        """
        super(FilerDoer, self).__init__(**kwa)
        self.filer = filer


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any
        """
        # inject temp into file resources here if any
        if not self.filer.opened:
            self.filer.reopen(temp=temp)


    def exit(self):
        """"""
        self.filer.close(clear=self.filer.temp)



# -*- encoding: utf-8 -*-
"""
hio.core.multidoing Module

Provides multiprocessing support



"""
import os
import time
import logging
import json
import multiprocessing as mp

from collections import deque, namedtuple
from dataclasses import dataclass, astuple, asdict, field


from . import tyming
from .doing import Doist, Doer
from .. import hioing
from .. import help
from ..help import timing, helping
from ..help.helping import RawDom
from ..help import ogling

from ..core.uxd import PeerMemoer

ogler = ogling.initOgler(prefix='hio_mp', name="Boss", level=logging.ERROR)


"""Bossage (namedtuple):
BossDoer info to be injected in CrewDoer as BossDoer enter time
Fields:
   name (str | None): name of Boss for resource management
   path (str | None): UXD cmd memo path used by crew to talk to their boss
"""
Bossage = namedtuple("Bossage", "name path")

"""Loadage (namedtuple):
BossDoer info to be injected into CrewDoer .start() containing both crew doist
parms for Process target kwargs and and CrewDoer parms

Fields:
   name (str): child doist identifier for resources.
   tyme (float): child doist start tyme
   tock (float | None): child doist tock, tyme lag between runs
   real (bool): child doist True means run in real time, tyme is real time
   limit (float | None): child doist tyme limit. None means run forever
   doers (list[Doers]): child doist List of Doers
   temp (bool | None): True means use temporary file resources
   boss (Bossage | None): BossDoer  info for CrewDoer. May be filled at enter time
                      CrewDoer uses to contact BossDoer.
"""
Loadage = namedtuple("Loadage", "name tyme tock real limit doers temp boss")


@dataclass
class CrewDom(RawDom):
    """Configuration dataclass of crew subprocess Doist parameters for crew doist
    and its CrewDoer to be injected by BossDoer.
    Use this when storing configuration in database or file. Use RawDom
    serialization hidden methods:

    Inherited Class Methods:
        _fromdict(cls, d: dict): return dataclass converted from dict d
        __iter__(self): asdict(self)
        _asdict(self): return self converted to dict
        _asjson(self): return self converted to json
        _ascbor(self): return self converted to cbor
        _asmgpk(self): return self converted to mgpk

    Attributes:
        name (str): child doist identifier for resources.
        tyme (float): child doist start tyme
        tock (float | None): child doist tock, tyme lag between runs
        real (bool): child doist True means run in real time, tyme is real time
        limit (float | None): child doist tyme limit. None means run forever
        doers (list[Doers]): child doist List of Doers
        temp (bool | None): True means use temporary file resources
        boss (Bossage | None): BossDoer  info for CrewDoer. May be filled at enter time
                                CrewDoer uses to contact BossDoer.
    """
    name: str ='child'  # unique identifier of child process and associated resources
    tyme: float = 0.0    # child doist start tyme
    tock: float | None = None  # child doist tyme lag between runs, None means ASAP
    real: bool = True  # child doist True means run in real time, tyme is real time
    limit: float | None = None  # child doist tyme limit. None mean run forever
    doers: list = field(default_factory=list)  # child doist list of doers
    temp: bool | None = False  # use temporary file resources if any
    boss: (Bossage | None) = Bossage(name=None, path=None)  # BossDoer info


@dataclass
class MemoDom(RawDom):
    """Inter Boss Crew Hand structured memo dataclass. Used for control messages
    Between Boss and Crew Doers via their .peer UXD BossMemoer or CrewMemoer.


    Inherited Class Methods:
        _fromdict(cls, d: dict): return dataclass converted from dict d
        __iter__(self): asdict(self)
        _asdict(self): return self converted to dict
        _asjson(self): return self converted to json
        _ascbor(self): return self converted to cbor
        _asmgpk(self): return self converted to mgpk

    Attributes:
        name (str): unique identifier as source of memo
        kin (str): type of memo
        load (dict): type specific payload of memo
    """
    name: str ='child'  # unique identifier of source
    kin: str = 'ACK'    # type of memo
    load: dict = field(default_factory=dict)  # type specific payload


class BossDoer(Doer):
    """BossDoer spawns multiple crew hand subprocesses and injects each with
    a Doist and Doers. The boss Doists runs the BossDoer in the parent process.
    Each crew hand Doist runs a CrewDoer that coordinates with the BossDoer.

    Analogy Boss runs a Crew of Hans. The parent process has a boss Doist which
    runs a BossDoer. Each crew hand is a child process with its own crew doist
    that runs its own CrewDoer

    See Doer for inherited attributes, properties, and methods.

    Inherited Attributes:
        done (bool): completion state:
                     True means completed fully. Otherwise incomplete.
                     Incompletion value may be None or False.
        opts (dict): schedulaer injects options from .opts into its .do generator
                     function as **opts parameter.


    Inherited Properties:
        tyme (float): is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        tymth (closure): function wrapper closure returned by Tymist.tymen()
                        method. When .tymth is called it returns associated
                        Tymist.tyme. Provides injected dependency on Tymist
                        tyme base.
        tock (float): desired time in seconds between runs or until next run,
                 non negative, zero means run asap


    Inherited Methods:
        __call__()  makes instance callable as generator function returning generator
        do() generator method that returns generator
        enter() is enter context action method
        recur() recur context action method or generator method
        clean() clean context action method
        exit() exit context method
        close() close context method
        abort() abort context method
        wind()  injects ._tymth dependency from associated Tymist to get its .tyme


    Attributes:
        name (str): unique identifier for this MultDoer boss
                    used to manage local resources
        peer (BossMemoer): underlying transport instance subclass of Memoer
        loads (list[dict]): BossDoer info to be injected into CrewDoer .start()
                            containing both crew doist parmss for Process target
                            kwargs and and CrewDoer parms
                            (see Loadage._asdict() or CrewDom._asdict())
        ctx (mp.context.SpawnContext | None): context under which to spawn processes
        crew (dict): values are child Process instances keyed by name



    Properties:


    """

    def __init__(self, *, name='boss', peer=None, loads=None, temp=False, **kwa):
        """Initialize instance.

        Inherited Parameters:
            tymth (closure): injected function wrapper closure returned by
                Tymist.tymen() instance. Calling tymth() returns associated
                Tymist.tyme.
            tock (float): seconds initial value of .tock
            opts (dict): injected options into its .do generator by scheduler

        Parameters:
            name (str): unique identifier for this BossDoer boss to be used
                        to manage local resources
            peer (BossMemoer | None): underlying transport instance subclass of Memoer
            loads (list[dict]): parameters used to spinup crew hand subprocess
                                .start(). See fields of Loadage and Bossage
            temp (bool): True means logger or other file resources created by
                            .start() will use temp
                         False otherwise


        """
        super(BossDoer, self).__init__(**kwa)
        self.name = name
        if not peer:
            peer = BossMemoer(name=self.name, reopen=False)  # default not opened reopen in enter
        self.peer = peer
        self.loads = loads if loads is not None else []
        self.temp = temp

        self.ctx = mp.get_context('spawn')
        self.crew = {}



    def enter(self, *, temp=None):
        """Do 'enter' context.
        Set up resources. Comparable to context manager enter.
        Start processes with config from .loads
        Not a generator method.


        Parameters:
            temp (bool | None): True means use temporary file resources if any.
                                None means ignore parameter value. Use self.temp.

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter

        """
        logger = ogler.getLogger()  # get so picks up test ogler.level
        logger.debug("BossDoer Enter: name=%s size=%d, ppid=%d, pid=%d, module=%s, temp=%s,ogler=%s.",
            self.name, len(self.loads), os.getppid(), os.getpid(), __name__, temp, ogler.name)
        self.peer.reopen(temp=temp)
        logger.debug("Boss name=%s Peer: name=%s path=%s opened=%s.",
                    self.name, self.peer.name, self.peer.path, self.peer.opened)
        boss = Bossage(name=self.peer.name, path=self.peer.path)  # use peer at enter

        for load in self.loads:
            #logger.debug("Load: %s.", load)
            name = load["name"]
            doers = load["doers"]
            if not doers or not isinstance(doers[0], CrewDoer):  # first doer must be CrewDoer
                raise hioing.MultiError(f"No crew doer for crew hand doist={name}.")

            doers[0].boss = boss  # assign boss contact info to crew doer

            if name not in self.crew:  # ensure unique by name
                hand = self.ctx.Process(name=name,
                                         target=self.start,
                                         kwargs=load)
                self.crew[hand.name] = hand
                hand.start()
            else:
                raise hioing.MultiError(f"Non-unique crew hand {name=} in loads.")


    def recur(self, tyme):
        """Do 'recur' context."""
        logger = ogler.getLogger()  # get so picks up test ogler.level
        self.peer.service()
        if not self.ctx.active_children():
            return True   # complete
        return False  # incomplete recur again


    def exit(self):
        """Do 'exit' (try finally) context."""
        logger = ogler.getLogger()
        self.peer.close(clear=True)
        logger.debug("Boss name=%s Peer: name=%s path=%s opened=%s.",
                self.name, self.peer.name, self.peer.path, self.peer.opened)
        logger.debug("BossDoer Exit: name=%s, ppid=%d, pid=%d, module=%s, ogler=%s.",
                self.name, os.getppid(), os.getpid(), __name__, ogler.name)


    def close(self):
        """Do 'close' context."""
        logger = ogler.getLogger()  # get so picks up test ogler.level


    def abort(self, ex):
        """Do 'abort' context."""
        logger = ogler.getLogger()  # get so picks up test ogler.level


    @staticmethod
    def start(*, name='crew', tyme=0.0, tock=None, real=True, limit=None,
               doers=None, temp=None, boss=None):
        """Process target function to spinup and run doist inside crew subprocess
        after it has been started.

        Parameters:
            name (str): unique crew hand name to be used to manage resources
            tyme (float): crew doist initial value of cycle time in seconds
            tock (float | None): crew doist tock time in seconds. None means run Asap
            real (bool): crew doist True means run in real time,
                        Otherwise run faster than real
            limit (float | None): crew doist seconds for max run time of doist.
                                  None means no limit.
            doers (iterable[Doer] | None): crew doist Doer class instances
                                   First entry must be CrewDoer
            temp (bool | None): True means use temp file resources by injection.
                                Otherwise ignore do not inject.
            boss (Bossage | None): boss info. May be filled at enter time
                                  CrewDoer uses to contact BossDoer.


        Doist must be built after process started so local tymth closure is created
        inside subprocess so that when doist winds the tyme for its doers
        the tymth closure is locally sourced.

        When run inside subprocess, this static method provides the outside
        scope for any Doist and doers that reference objects inside the
        __main__ moduleo that calls this method such as ogler.
        The subprocess gets a picked copy of these module globals. Any Doer
        inside then has access to the module globals and can update them as
        neede. In the case of ogler this means changing ogler.level, ogler.temp
        and running ogler.reopen(temp=temp) as appropriate.
        """
        logger = ogler.getLogger()  # do here so picks up level inside subprocess

        logger.debug("Crew Start: name=%s, ppid=%d, pid=%s, module=%s, temp=%s, ogler=%s.",
                        name, os.getppid(), os.getpid(), __name__, temp, ogler.name)
        time.sleep(0.01)

        doist = Doist(name=name, tyme=tyme, tock=tock, real=real, limit=limit,
                      doers=doers, temp=temp)
        doist.do()

        logger.debug("Crew End: name=%s, ppid=%d, pid=%s, module=%s, temp=%s, ogler=%s.",
                        name, os.getppid(), os.getpid(), __name__, temp, ogler.name)


class CrewDoer(Doer):
    """CrewDoer runs interface between a given crew hand subprocess and its
    boss process. This must be first doer run by crew hand subprocess doist.

    Attributes:
        name (str): crew hand name for managing local resources to subprocess
        peer (CrewMemoer): underlying transport instance subclass of Memoer
        count (int): iteration counter for debugging

    """

    def __init__(self, *, name='crew', boss=Bossage(name=None, path=None),
                          peer=None, **kwa):
        """Initialize instance.

        Inherited Parameters:
            tymth (closure): injected function wrapper closure returned by
                Tymist.tymen() instance. Calling tymth() returns associated
                Tymist.tyme.
            tock (float): seconds initial value of .tock
            opts (dict): injected options into its .do generator by scheduler

        Parameters:
            name (str): unique identifier for this BossDoer boss to be used
                        to manage local resources
            boss (Bossage): contact info for BossDoer. assigned by boss at enter
            peer (CrewMemoer | None): underlying transport instance subclass of Memoer

            temp (bool): True means logger or other file resources created by
                            .start() will use temp
                         False otherwise
        """
        super(CrewDoer, self).__init__(**kwa)
        self.name = name
        self.boss = boss
        self.peer = peer if peer is not None else CrewMemoer(name=self.name, reopen=False)
        self.count = None


    def enter(self, *, temp=None):
        """Do 'enter' context.
        Set up resources. Comparable to context manager enter.
        Not a generator method.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        logger = ogler.getLogger()  # get so picks up test ogler.level
        self.count = 0
        logger.debug("CrewDoer Enter: name=%s pid=%d, temp=%s, ogler=%s, count=%d.",
                    self.name, os.getpid(), temp, ogler.name, self.count)
        logger.debug("Crew name=%s Boss: name=%s path=%s.",
                            self.name, self.boss.name, self.boss.path)
        self.peer.reopen(temp=temp)
        logger.debug("Crew name=%s Peer: name=%s path=%s opened=%s.",
                    self.name, self.peer.name, self.peer.path, self.peer.opened)

        memo = dict(name=self.name, kin="REG", load={})
        memo = json.dumps(memo, separators=(",", ":"), ensure_ascii=False)
        dst = self.boss.path
        self.peer.memoit(memo, dst)


    def recur(self, tyme):
        """Do 'recur' context."""
        logger = ogler.getLogger()
        self.count += 1
        logger.debug("CrewDoer Recur: name=%s, pid=%d, ogler=%s, count=%d.",
                    self.name, os.getpid(), ogler.name, self.count)
        self.peer.service()
        if self.count > 3:
            return True  # complete
        return False  # incomplete


    def exit(self):
        """Do 'exit' (try finally) context."""
        logger = ogler.getLogger()
        self.count += 1
        self.peer.close(clear=True)
        logger.debug("Crew name=%s Peer: name=%s path=%s opened=%s.",
                    self.name, self.peer.name, self.peer.path, self.peer.opened)
        logger.debug("CrewDoer Exit: name=%s pid=%d, ogler=%s, count=%d.",
                    self.name, os.getpid(), ogler.name, self.count)

    def close(self):
        """Do 'close' context."""
        logger = ogler.getLogger()
        self.count += 1


    def abort(self, ex):
        """Do 'abort' context."""
        logger = ogler.getLogger()
        self.count += 1



class BossMemoer(PeerMemoer):
    """Class for sending memograms over UXD transport
    Mixin base classes Peer and Memoer to attain memogram over uxd transport.


    Inherited Class Attributes:
        MaxGramSize (int): absolute max gram size on tx with overhead
        See memoing.Memoer Class
        See Peer Class

    Inherited Attributes:
        See memoing.Memoer Class
        See Peer Class

    Class Attributes:

    Attributes:

    """


    def __init__(self, *, reopen=False, bc=4, **kwa):
        """Initialization method for instance.

        Inherited Parameters:
            reopen (bool): True (re)open with this init
                           False not (re)open with this init but later (default)
            bc (int | None): count of transport buffers of MaxGramSize

            See memoing.Memoer for other inherited paramters
            See Peer for other inherited paramters


        Parameters:

        """
        super(BossMemoer, self).__init__(reopen=reopen, bc=bc, **kwa)


    def serviceRxMemos(self):
        """Service all memos in .rxms (greedy) if any

        Override in subclass to handle result(s) and put them somewhere
        """
        logger = ogler.getLogger()

        while self.rxms:
            memo, src, vid = self._serviceOneRxMemo()
            logger.debug("Boss Peer RX: name=%s rx from src=%s memo=%s.",
                                self.name, src, memo)
            memo = json.loads(memo)
            if memo['kin'] == "REG":
                pass # register src path with name


            dst = src
            mack = dict(name=self.name, kin="ACK", load={})
            mack = json.dumps(mack,separators=(",", ":"),ensure_ascii=False)
            self.memoit(mack, dst)



class CrewMemoer(PeerMemoer):
    """Class for sending memograms over UXD transport
    Mixin base classes Peer and Memoer to attain memogram over uxd transport.


    Inherited Class Attributes:
        MaxGramSize (int): absolute max gram size on tx with overhead
        See memoing.Memoer Class
        See Peer Class

    Inherited Attributes:
        See memoing.Memoer Class
        See Peer Class

    Class Attributes:

    Attributes:

    """
    def __init__(self, *, reopen=False, bc=4, **kwa):
        """Initialization method for instance.

        Inherited Parameters:
            reopen (bool): True (re)open with this init
                           False not (re)open with this init but later (default)
            bc (int | None): count of transport buffers of MaxGramSize

            See memoing.Memoer for other inherited paramters
            See Peer for other inherited paramters


        Parameters:

        """
        super(CrewMemoer, self).__init__(reopen=reopen, bc=bc, **kwa)

    def serviceRxMemos(self):
        """Service all memos in .rxms (greedy) if any

        Override in subclass to handle result(s) and put them somewhere
        """
        logger = ogler.getLogger()

        while self.rxms:
            memo, src, vid = self._serviceOneRxMemo()
            logger.debug("Crew Peer RX: name=%s rx from src=%s memo=%s.",
                                self.name, src, memo)
            memo = json.loads(memo)



# -*- encoding: utf-8 -*-
"""
hio.core.uxd.peermemoing Module
"""
from contextlib import contextmanager

from ... import help
from ... import hioing
from ...base import doing
from ..uxd import Peer
from ..memo import Memoer, GramDex

logger = help.ogler.getLogger()


class PeerMemoer(Peer, Memoer):
    """Class for sending memograms over UXD transport
    Mixin base classes Peer and Memoer to attain memogram over uxd transport.


    Inherited Class Attributes:
        MaxGramSize (int): absolute max gram size on tx with overhead
        See memoing.Memoer Class
        See Peer Class

    Inherited Attributes:
        See memoing.Memoer Class
        See Peer Class

    Class Attributes:


    Attributes:


    """


    def __init__(self, *, bc=4, **kwa):
        """Initialization method for instance.

        Inherited Parameters:
            bc (int | None): count of transport buffers of MaxGramSize

            See memoing.Memoer for other inherited paramters
            See Peer for other inherited paramters


        Parameters:

        """
        super(PeerMemoer, self).__init__(bc=bc, **kwa)



@contextmanager
def openPM(cls=None, name="test", temp=True, reopen=True, clear=True,
             filed=False, extensioned=True, **kwa):
    """
    Wrapper to create and open UXD PeerMemoer instances
    When used in with statement block, calls .close() on exit of with block

    Parameters:
        cls (Class): instance of subclass instance
        name (str): unique identifer of PeerMemoer peer.
                    Enables management of transport by name.
                    Provides unique path part so can have many peers each at
                    different paths but in same directory.
        temp (bool): True means open in temporary directory, clear on close
                     Otherwise open in persistent directory, do not clear on close
        reopen (bool): True (re)open with this init
                       False not (re)open with this init but later (default)
        clear (bool): True means remove directory upon close when reopening
                      False means do not remove directory upon close when reopening
        filed (bool): True means .path is file path not directory path
                      False means .path is directiory path not file path
        extensioned (bool): When not filed:
                            True means ensure .path ends with fext
                            False means do not ensure .path ends with fext

    See filing.Filer and uxding.Peer for other keyword parameter passthroughs

    Usage:
        with openPM() as peer:
            peer.receive()

        with openPM(cls=PeerMemoerSubclass) as peer:
            peer.receive()

    """
    peer = None
    if cls is None:
        cls = PeerMemoer
    try:
        peer = cls(name=name, temp=temp, reopen=reopen, clear=clear,
                   filed=filed, extensioned=extensioned, **kwa)

        yield peer

    finally:
        if peer:
            peer.close(clear=peer.temp or clear)



class PeerMemoerDoer(doing.Doer):
    """PeerMemoerDoer Doer for reliable UXD transport.
    Does not require retry tymers.

    See Doer for inherited attributes, properties, and methods.
    To test in WingIde must configure Debug I/O to use external console

    Attributes:
       .peer (PeerMemoerDoer): underlying transport instance subclass of Memoer

    """

    def __init__(self, peer, **kwa):
        """Initialize instance.

        Parameters:
           peer (Peer): is Memoer Subclass instance
        """
        super(PeerMemoerDoer, self).__init__(**kwa)
        self.peer = peer


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        # inject temp into file resources here if any
        self.peer.reopen(temp=temp)


    def recur(self, tyme):
        """"""
        self.peer.service()


    def exit(self):
        """"""
        self.peer.close(clear=True)




"""
Asynchronous (nonblocking) serial io

"""
from __future__ import absolute_import, division, print_function

import sys
import os
import errno
import platform
from collections import deque


from ... import hioing, help
from ...base import doing

logger = help.ogler.getLogger()


class LineError(hioing.HioError):
    """
    Serial line error. Too big for buffer.

    Usage:
        raise LineError("error message")
    """

class Console():
    """
    Class to manage non blocking interactive I/O on serial console

    Opens non blocking read file descriptor on console
    Use instance method .close to close file descriptor
    Use instance methods .getline to read & .put to write to console
    Needs os module

    Attributes:
        bs (int): max buffer size for each read, defaults to 256
        fd (int):  file descriptor for console
        opened (bool): True means .fd opened, False means .fd closed
        rxbs (bytearray): of received characters (bytes)

    Methods:
        reopen ():  closes and reopens .fd, sets .opened
        close ():   closes .fd unsets .opened
        get (): returns chars including newline but no more than bs characters
        put ():  puts characters

    Hidden:
        ._line is bytearray of line buffer

    """
    MaxBufSize = 256

    def __init__(self, bs=None):
        """
        Initialization method for instance.
        Creates attributes.
        """
        self.bs = bs if bs is not None else self.MaxBufSize
        self.fd = None  # console file descriptor needs to be opened
        self.opened =  False
        self.rxbs = bytearray()


    def open(self, port=''):
        """
        Opens fd on terminal console in non blocking mode.

        port is the serial port device path name
        or if '' then use os.ctermid() which
        returns path name of console usually '/dev/tty'

        os.O_NONBLOCK makes non blocking io
        os.O_RDWR allows both read and write.
        os.O_NOCTTY don't make this the controlling terminal of the process
        O_NOCTTY is only for cross platform portability BSD never makes it the
        controlling terminal

        Don't use print at same time since it will mess up non blocking reads.

        Works in both canonical and non-canonical input mode.
        In canonical mode, no chars are available to read until eol newline
        is entered and eol is included in the read characters.

        It appears that canonical mode is the default for fd console os.ctermid().
        For other serial port fds  the characters may be  available immediately.

        To debug use os.isatty(fd) which returns True if the file descriptor
        fd is open and connected to a tty-like device, else False.

        On UNIX/macOS systems uses os.ctermid() to get console
        On Windows systems uses console directly via msvcrt
        """
        if not port:
            system = platform.system()
            if system == 'Windows':
                # Windows doesn't need a specific port for console access via msvcrt
                port = None
            else:
                # Unix/macOS use ctermid
                port = os.ctermid()  # default to console

        try:
            if platform.system() == 'Windows':
                # Windows handling through msvcrt
                import msvcrt
                # For Windows, we'll use stdin/stdout file descriptors
                # and rely on msvcrt for non-blocking operations
                self.fd = 0  # Use 0 as a placeholder value for Windows console
                # No specific open operation needed for Windows console via msvcrt
            else:
                # Unix/macOS handling
                self.fd = os.open(port, os.O_NONBLOCK | os.O_RDWR | os.O_NOCTTY)
        except OSError as ex:
            logger.error("Error opening console serial port, %s\n", ex)
            return False

        self.opened = True
        return self.opened


    def reopen(self, **kwa):
        """
        Idempotently opens console
        """
        self.close()
        return self.open()


    def close(self):
        """
        Closes fd.
        """
        if self.fd:
            os.close(self.fd)
        self.fd = None
        del self.rxbs[:]
        self.opened = False


    def put(self, data = b'\n'):
        """
        Writes data bytes to console and return number of bytes from data written.
        """
        if platform.system() == 'Windows':
            # Windows-specific console output
            import msvcrt
            # Write bytes to stdout
            for b in data:
                msvcrt.putch(bytes([b]))
            return len(data)
        else:
            # Unix/macOS
            return os.write(self.fd, data)  # returns number of bytes written


    def get(self, bs=None):
        """
        Gets nonblocking line of bytes from console of up to bs characters
        including eol newline if in bs characters otherwise
        must repeat get until a newline appears.

        Returns empty string if no characters available else returns line.
        Works in both canonical and non-canonical mode
        In canonical mode, no chars are available to read until eol newline
        is entered and eol is included in the read characters.

        Strips eol newline before returning line.
        """
        bs = bs if bs is not None else self.bs
        line = bytearray()

        try:
            if platform.system() == 'Windows':
                # Windows-specific non-blocking read using msvcrt
                import msvcrt
                # Check if any keys are available
                while msvcrt.kbhit():
                    # Read a character and add it to our buffer
                    char = msvcrt.getch()
                    self.rxbs.extend(char)
            else:
                # Unix/macOS non-blocking read
                self.rxbs.extend(os.read(self.fd, bs))
        except OSError as ex1:  # if no chars available generates exception
            # ex.args[0] == ex.errno for better os compatibility
            # the value of a given errno.XXXXX may be different on each os
            # EAGAIN: BSD 35, Linux 11, Windows 11
            # EWOULDBLOCK: BSD 35 Linux 11 Windows 140
            if ex1.args[0] in (errno.EAGAIN, errno.EWOULDBLOCK):
                pass  # No characters available
            else:
                logger.error("Error: Get on Console '%s'."
                              " '%s'\n", self.fd, ex1)
                raise  # re-raise exception ex1

        # Process any complete lines in the buffer
        if (idx := self.rxbs.find(ord(b'\n'))) != -1:
            line.extend(self.rxbs[:idx])  # copy all but newline
            del self.rxbs[:idx+1]  # delete including newline
        elif platform.system() == 'Windows' and (idx := self.rxbs.find(ord(b'\r'))) != -1:
            # On Windows, handle CR as line terminator too
            line.extend(self.rxbs[:idx])  # copy all but CR
            del self.rxbs[:idx+1]  # delete including CR

        return line


    def service(self):
        """
        Service puts and gets
        """


class ConsoleDoer(doing.Doer):
    """
    Basic Console Doer. Wraps console in doer context so opens and closes console

    To test in WingIde must configure Debug I/O to use external console
    See Doer for inherited attributes, properties, and methods.

    Attributes:
       .console is serial Console instance

    """

    def __init__(self, console, **kwa):
        """
        Initialize instance.

        Parameters:
           console is serial Console instance

        """
        super(ConsoleDoer, self).__init__(**kwa)
        self.console = console


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        # inject temp into file resources here if any
        result = self.console.reopen(temp=temp)


    def recur(self, tyme):
        """"""
        self.console.service()


    def exit(self):
        """"""
        self.console.close()


class EchoConsoleDoer(doing.Doer):
    """
    Basic Terminal Console IO to buffers. Echos input back to output

    To test in WingIde must configure Debug I/O to use external console

    See Doer for inherited attributes, properties, and methods.

    Attributes:
       .console is serial Console instance

    """

    def __init__(self, console, lines=None, txbs=None, **kwa):
        """
        Initialize instance.

        Parameters:
           console is serial Console instance
           lines is deque of input bytes bytearrays of each line from console
           txbs is ouput bytes bytearray to send to console

        """
        super(EchoConsoleDoer, self).__init__(**kwa)
        self.console = console
        self.lines = lines if lines is not None else deque()
        self.txbs = txbs if txbs is not None else bytearray()


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        # inject temp into file resources here if any
        self.console.reopen(temp=temp)
        self.txbs.extend(b"Cmds: q=quit, h=help otherwise echoes.\n")
        self.txbs.extend(b"Type cmd & \n: ")


    def recur(self, tyme):
        """"""
        done = False
        prompt = False
        while self.lines:
            line = self.lines.popleft()
            #process line here
            if line == b'q':
                self.txbs.extend(b"Goodbye\n.")
                done = True  #  all done so indicate exit
                break

            elif line == b'h':
                self.txbs.extend(b"Help: type q to quit or h for help.\n")

            else:
                self.txbs.extend(b"Echo: %s\n" % line )

            prompt = True

        if prompt:
            self.txbs.extend(b"Type cmd & \n: ")


        if self.txbs:
            count =  self.console.put(self.txbs)  #  write
            del self.txbs[:count]

        line = self.console.get()  #  read
        if line:
            self.lines.append(line)


        return done  # keep going if done == False else ends

    def exit(self):
        """"""
        self.console.close()



class Device():
    """
    Class to manage non blocking IO on serial device port.

    Opens non blocking read file descriptor on serial port
    Use instance method close to close file descriptor
    Use instance methods get & put to read & write to serial device
    Needs os module
    """
    def __init__(self, port=None, speed=9600, bs=1024):
        """
        Initialization method for instance.

        port = serial device port path string
        speed = serial port speed in bps
        bs = buffer size for reads
        """
        self.fd = None  # serial device port file descriptor, must be opened first
        self.port = port

        if not self.port:
            system = platform.system()
            if system == 'Windows':
                self.port = 'COM1'  # Default Windows serial port
            else:
                self.port = os.ctermid()  # Default to console on Unix/macOS

        self.speed = speed or 9600
        self.bs = bs or 1024
        self.opened = False


    def reopen(self, port=None, speed=None, bs=None):
        """
        Idempotently open serial device port
        Opens fd on serial port in non blocking mode.

        port is the serial port device path name or
        if '' then use os.ctermid() which
        returns path name of console usually '/dev/tty'

        os.O_NONBLOCK makes non blocking io
        os.O_RDWR allows both read and write.
        os.O_NOCTTY don't make this the controlling terminal of the process
        O_NOCTTY is only for cross platform portability BSD never makes it the
        controlling terminal

        Don't use print and console at same time since it will mess up non blocking reads.

        The input mode, canonical or noncanonical, is controlled by the
        ICANON flag see termios module.

        Raw mode

        def setraw(fd, when=TCSAFLUSH):
            Put terminal into a raw mode.
            mode = tcgetattr(fd)
            mode[IFLAG] = mode[IFLAG] & ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON)
            mode[OFLAG] = mode[OFLAG] & ~(OPOST)
            mode[CFLAG] = mode[CFLAG] & ~(CSIZE | PARENB)
            mode[CFLAG] = mode[CFLAG] | CS8
            mode[LFLAG] = mode[LFLAG] & ~(ECHO | ICANON | IEXTEN | ISIG)
            mode[CC][VMIN] = 1
            mode[CC][VTIME] = 0
            tcsetattr(fd, when, mode)


        # set up raw mode / no echo / binary
        cflag |=  (TERMIOS.CLOCAL|TERMIOS.CREAD)
        lflag &= ~(TERMIOS.ICANON|TERMIOS.ECHO|TERMIOS.ECHOE|TERMIOS.ECHOK|TERMIOS.ECHONL|
                     TERMIOS.ISIG|TERMIOS.IEXTEN) #|TERMIOS.ECHOPRT
        for flag in ('ECHOCTL', 'ECHOKE'): # netbsd workaround for Erk
            if hasattr(TERMIOS, flag):
                lflag &= ~getattr(TERMIOS, flag)

        oflag &= ~(TERMIOS.OPOST)
        iflag &= ~(TERMIOS.INLCR|TERMIOS.IGNCR|TERMIOS.ICRNL|TERMIOS.IGNBRK)
        if hasattr(TERMIOS, 'IUCLC'):
            iflag &= ~TERMIOS.IUCLC
        if hasattr(TERMIOS, 'PARMRK'):
            iflag &= ~TERMIOS.PARMRK

        """
        self.close()

        if port is not None:
            self.port = port
        if speed is not None:
            self.speed = speed
        if bs is not None:
            self.bs = bs

        system = platform.system()

        if system == 'Windows':
            try:
                # Try to use pyserial for COM ports on Windows for better handling
                import serial
                self._serial = serial.Serial(port=self.port,
                                        baudrate=self.speed,
                                        timeout=0)
                # Placeholder for fd
                self.fd = 0
            except ImportError:
                # Fall back to basic file operations if pyserial is not available
                self.fd = os.open(self.port, os.O_RDWR | os.O_BINARY)
                import msvcrt
                msvcrt.setmode(self.fd, os.O_BINARY)
                self._serial = None
        else:
            # Unix/macOS handling
            self.fd = os.open(self.port, os.O_NONBLOCK | os.O_RDWR | os.O_NOCTTY)
            self._serial = None

        if (system == 'Darwin') or (system == 'Linux'):  # use termios to set values
            import termios

            iflag, oflag, cflag, lflag, ispeed, ospeed, cc = range(7)

            settings = termios.tcgetattr(self.fd)
            #print(settings)

            settings[lflag] = (settings[lflag] & ~termios.ICANON)

            settings[lflag] = (settings[lflag] & ~termios.ECHO) # no echo

            #ignore carriage returns on input
            #settings[iflag] = (settings[iflag] | (termios.IGNCR)) #ignore cr

            # 8N1 8bit word no parity one stop bit nohardware handshake ctsrts
            # to set size have to mask out(clear) CSIZE bits and or in size
            settings[cflag] = ((settings[cflag] & ~termios.CSIZE) | termios.CS8)
            # no parity clear PARENB
            settings[cflag] = (settings[cflag] & ~termios.PARENB)
            #one stop bit clear CSTOPB
            settings[cflag] = (settings[cflag] & ~termios.CSTOPB)
            #no hardware handshake clear crtscts
            settings[cflag] = (settings[cflag] & ~termios.CRTSCTS)

            # in linux the speed flag does not equal value so always set it
            speedattr = "B{0}".format(self.speed)  # convert numeric speed to attribute name string
            speed = getattr(termios, speedattr)
            settings[ispeed] = speed
            settings[ospeed] = speed

            termios.tcsetattr(self.fd, termios.TCSANOW, settings)
            #print(settings)

        self.opened = True

        return self.opened



    def close(self):
        """
        Closes fd.
        """
        if self._serial:
            self._serial.close()
            self._serial = None

        if self.fd and platform.system() != 'Windows' or not self._serial:
            # Close fd if not Windows or if we didn't use pyserial
            os.close(self.fd)

        self.fd = None
        self.opened = False


    def receive(self):
        """
        Reads nonblocking characters from serial device up to bs characters
        Returns empty bytes if no characters available else returns all available.
        In canonical mode no chars are available until newline is entered.
        """
        data = b''
        try:
            if platform.system() == 'Windows':
                if self._serial:
                    # Use pyserial if available
                    data = self._serial.read(self.bs)
                else:
                    # Fall back to msvcrt for console
                    import msvcrt
                    if msvcrt.kbhit():
                        char = msvcrt.getch()
                        data = char
                    else:
                        data = os.read(self.fd, self.bs)
            else:
                # Unix/macOS read
                data = os.read(self.fd, self.bs)  #if no chars available generates exception
        except OSError as ex1:  # ex1 is the target instance of the exception
            # ex.args[0] == ex.errno for better os compatibility
            # the value of a given errno.XXXXX may be different on each os
            # EAGAIN: BSD 35, Linux 11, Windows 11
            # EWOULDBLOCK: BSD 35 Linux 11 Windows 140
            if ex1.args[0] in (errno.EAGAIN, errno.EWOULDBLOCK):
                pass #No characters available
            else:
                logger.error("Error: Receive on Device '%s'."
                              " '%s'\n", self.port, ex1)
                raise #re raise exception ex1

        return data

    def send(self, data=b'\n'):
        """
        Writes data bytes to serial device port.
        Returns number of bytes sent
        """
        try:
            if platform.system() == 'Windows' and self._serial:
                # Use pyserial if available
                count = self._serial.write(data)
            else:
                count = os.write(self.fd, data)
        except OSError as ex1:  # ex1 is the target instance of the exception
            # ex.args[0] == ex.errno for better os compatibility
            # the value of a given errno.XXXXX may be different on each os
            # EAGAIN: BSD 35, Linux 11, Windows 11
            # EWOULDBLOCK: BSD 35 Linux 11 Windows 140
            if ex1.args[0] in (errno.EAGAIN, errno.EWOULDBLOCK):
                count = 0  # buffer full can't write
            else:
                logger.error("Error: Send on Device '%s'."
                              " '%s'\n", self.port, ex1)
                raise #re raise exception ex1

        return count


class Serial():
    """
    Class to manage non blocking IO on serial device port using pyserial

    Opens non blocking read file descriptor on serial port
    Use instance method close to close file descriptor
    Use instance methods get & put to read & write to serial device
    Needs os module
    """
    def __init__(self, port=None, speed=9600, bs=1024):
        """
        Initialization method for instance.

        port = serial device port path string
        speed = serial port speed in bps
        bs = buffer size for reads


        """
        self.serial = None  # Serial instance
        self.port = port

        if not self.port:
            system = platform.system()
            if system == 'Windows':
                self.port = 'COM1'  # Default Windows serial port
            else:
                self.port = os.ctermid()  # Default to console on Unix/macOS

        self.speed = speed or 9600
        self.bs = bs or 1024
        self.opened = False


    def reopen(self, port=None, speed=None, bs=None):
        """
        Opens fd on serial port in non blocking mode.

        port is the serial port device path name or
        if None then use os.ctermid() which returns path name of console
        usually '/dev/tty'
        """
        self.close()

        if port is not None:
            self.port = port
        if speed is not None:
            self.speed = speed
        if bs is not None:
            self.bs = bs

        import serial  # import pyserial
        self.serial = serial.Serial(port=self.port,
                                    baudrate=self.speed,
                                    timeout=0,
                                    writeTimeout=0)
        #self.serial.nonblocking()
        self.serial.reset_input_buffer()
        self.opened = True

        return self.opened


    def close(self):
        """
        Closes .serial
        """
        if self.serial:
            self.serial.reset_output_buffer()
            self.serial.close()
            self.serial = None
            self.opened = False

    def receive(self):
        """
        Reads nonblocking characters from serial device up to bs characters
        Returns empty bytes if no characters available else returns all available.
        In canonical mode no chars are available until newline is entered.
        """
        data = b''
        try:
            data = self.serial.read(self.bs)  #if no chars available generates exception
        except OSError as ex1:  # ex1 is the target instance of the exception
            # ex.args[0] == ex.errno for better os compatibility
            # the value of a given errno.XXXXX may be different on each os
            # EAGAIN: BSD 35, Linux 11, Windows 11
            # EWOULDBLOCK: BSD 35 Linux 11 Windows 140
            if ex1.args[0] in (errno.EAGAIN, errno.EWOULDBLOCK):
                pass #No characters available
            else:
                logger.error("Error: Receive on Serial '%s'."
                              " '%s'\n", self.port, ex1)
                raise #re raise exception ex1

        return data

    def send(self, data=b'\n'):
        """
        Writes data bytes to serial device port.
        Returns number of bytes sent
        """
        try:
            count = self.serial.write(data)
        except OSError as ex1:  # ex1 is the target instance of the exception
            # ex.args[0] == ex.errno for better os compatibility
            # the value of a given errno.XXXXX may be different on each os
            # EAGAIN: BSD 35, Linux 11, Windows 11
            # EWOULDBLOCK: BSD 35 Linux 11 Windows 140
            if ex1.args[0] in (errno.EAGAIN, errno.EWOULDBLOCK):
                count = 0  # buffer full can't write
            else:
                logger.error("Error: Send on Serial '%s'."
                              " '%s'\n", self.port, ex1)
                raise #re raise exception ex1

        return count



class Driver():
    """
    Nonblocking Serial Device Port Driver
    """

    def __init__(self,
                 name=u'',
                 uid=0,
                 port=None,
                 speed=9600,
                 bs=1024,
                 server=None):
        """
        Initialization method for instance.

        Parameters:
            name = user friendly name for driver
            uid = unique identifier for driver
            port = serial device port path string
            speed = serial port speed in bps
            canonical = canonical mode True or False
            bs = buffer size for reads
            server = serial port device server if any

        Attributes:
           name = user friendly name for driver
           uid = unique identifier for driver
           server = serial device server nonblocking
           txbs = bytearray of data bytes to send
           rxbs = bytearray of data bytes received

        """
        self.name = name
        self.uid = uid

        if not server:
            try:
                import serial
                self.server = Serial(port=port,
                                       speed=speed,
                                       bs=bs)

            except ImportError as  ex:
                logger.error("Error: importing pyserial\n%s\n", ex)
                self.server = Device(port=port,
                                       speed=speed,
                                       bs=bs)
        else:
            self.server = server

        self.txbs = bytearray()  # bytearray of data to send
        self.rxbs = bytearray()  # bytearray of data received


    def serviceReceives(self):
        """
        Service receives until no more
        """
        while self.server.opened:
            data = self.server.receive()  # bytes
            if not data:
                break
            self.rxbs.extend(data)


    def clearRxbs(self):
        """
        Clear .rxbs
        """
        del self.rxbs[:]


    def scan(self, start):
        """
        Returns offset of given start byte in self.rxbs
        Returns None if start is not given or not found
        If strip then remove any bytes before offset
        """
        offset = self.rxbs.find(start)
        if offset < 0:
            return None
        return offset


    def send(self, data):
        """
        Handle one tx data
        """
        count = self.server.send(data)
        return count


    def tx(self, data):
        """
        Queue data onto .txbs
        """
        self.txbs.extend(data)


    def serviceSends(self):
        """
        Service .txbs
        """
        while self.txbs and self.server.opened:
            count = self.send(self.txbs)
            del self.txbs[:count]
            break  # try again later


    def service(self):
        """
        Sevice receives and sends
        """
        self.serviceReceives()
        self.serviceSends()





# -*- encoding: utf-8 -*-
"""
hio.core.memo Package
"""


from .memoing import (Versionage, Sizage, GramDex, SGDex,
                      openMemoer, Memoer, MemoerDoer,
                      openTM, TymeeMemoer, TymeeMemoerDoer)

